/*
 * Luigi Findanno 12/2019
 * Tachimeter/Accelerometer GPS
 * 
 * - Arduino Leonardo
 * - 2.8" display
 * - NMEA GPS Module 9600 bps
 *
 * Power the GPS Module and connect the TX pin to the 0(rx) pin of Arduino Leonardo 
 */

#include <Adafruit_GFX.h>    // Core graphics library
#include <Adafruit_TFTLCD.h> // Hardware-specific library

// The control pins for the LCD can be assigned to any digital or
// analog pins...but we'll use the analog pins as this allows us to
// double up the pins with the touch screen (see the TFT paint example).
#define LCD_CS A3 // Chip Select goes to Analog 3
#define LCD_CD A2 // Command/Data goes to Analog 2
#define LCD_WR A1 // LCD Write goes to Analog 1
#define LCD_RD A0 // LCD Read goes to Analog 0

#define LCD_RESET A4 // Can alternately just connect to Arduino's reset pin


#define BLACK   0x0000
#define BLUE    0x001F
#define RED     0xF800
#define GREEN   0x07E0
#define CYAN    0x07FF
#define MAGENTA 0xF81F
#define YELLOW  0xFFE0
#define WHITE   0xFFFF

char NmeaSentence[255];
int ni=0;
char GpsStatus = 'V', DispGpsStatus = 'A', Direc = 'S', MaxGForcDir = 'S';

float Vel = 0, MaxVel = 0, DispVel = 1, DispMaxVel = 1;
float GForc = 1, MaxGForc = 1, DispMaxGForc = 0, DispGForc = 0;
float Accel = 0, MaxAccel = 0, DispAccel = 1, MinAccel=0, DispMinAccel=1, DispMaxAccel = 1;

int CalcSwitch = 0;

Adafruit_TFTLCD tft(LCD_CS, LCD_CD, LCD_WR, LCD_RD, LCD_RESET);

void setup() {
  Serial1.begin(9600); // baud rate of GPS module
  Serial.begin(115200); // redirect the GPS data to debug serial

  tft.reset();

  uint16_t identifier = tft.readID();
  identifier=0x9341;
  
  tft.begin(identifier);
  tft.setRotation(3);


  InitDisplay();
}

void loop() {
  
  if (Serial1.available()){
    char inByte = Serial1.read();
    Serial.write(inByte);
    
    if (inByte == 13 || inByte == 10){ // end sentence
      NmeaSentence[ni] = 0;
      if (ni > 0) CheckNmea(); // check the data only one time every end of string
    }else{
      NmeaSentence[ni] = inByte;
      ni++;  
    }  
  }
}


void CheckNmea(){


  if (NmeaSentence[3] == 'V' && NmeaSentence[4] == 'T' && NmeaSentence[5] == 'G'){ // speed and acceleration

    char *vel;
    int c = 0;
    
    int NmeaLen, vellen;
    NmeaLen = strlen(NmeaSentence);
    
    for(int k = 0; k < NmeaLen; k++){
      if (NmeaSentence[k] == ',') c++;
      if (c == 7){ // after the 7th comma there is the speed data
        /*
         * Extract data comma separated.
         * When GPS is not active the string contain no numbers, there are no
         * characters between the commas ",,".
         * strtok doesn't work as I want, ",," double comma count as one.
         * In case of ",," I fill the zero value directly in the NmeaSentence array.
         */
        if (NmeaSentence[k + 1] == ','){ 
          NmeaSentence[k + 1] = '0';
          NmeaSentence[k + 2] = '.';
          NmeaSentence[k + 3] = '0';
          NmeaSentence[k + 4] = ',';
          
        }
          
        vel = NmeaSentence + k + 1; // vel point to the first character of the speed value 
        vellen = strlen(vel);
        
        for(int f = 0; f < vellen; f++){
          if (*(vel + f) == ',') {  // look for the comma after the value
            *(vel + f) = 0;         // change the comma with end of string character (zero)
            
            // values
            Vel = atof(vel);               // speed Km/h
            Accel = (Vel - DispVel) / 3.6; // acceleration m/s2
            GForc = sqrt(1 + pow((Accel / 9.8), 2)); // g force calculation

            /*
             * To prevent errors, the peak values calculation routine will be enabled
             * only if: 
             *  1) the GPS status is active 
             *  2) the acceleration value is < 0.28 m/s2 and >=-0.28 m/s2
             * When the GPS status is void, peak values calculation is disabled
             */

            if ((GpsStatus == 'A') && (Accel<=0.28) && (Accel>=-0.28)) CalcSwitch = 1;
            if (GpsStatus == 'V') CalcSwitch = 0;

            if (CalcSwitch == 1){ 
              // Peak values calculation
              if (Vel > MaxVel) MaxVel = Vel;
                
              if (Accel > MaxAccel) MaxAccel = Accel;
              if (Accel < MinAccel) MinAccel = Accel;
              if (Accel > 0) Direc = 'A'; else if (Accel < 0) Direc = 'D'; else if (Accel == 0) Direc = 'S';
                             
              if (GForc > MaxGForc){
                MaxGForc = GForc;
                MaxGForcDir = Direc;
              }
              
            }
            break;              
          }
        }
        break;          
      }
    }
  }

  if (NmeaSentence[3] == 'R' && NmeaSentence[4] == 'M' && NmeaSentence[5] == 'C'){ // GPS Active-Void
    int c = 0;
    for(int k = 0; k < strlen(NmeaSentence); k++){
      if (NmeaSentence[k] == ',') c++;
      if (c == 2){ // after the 2nd comma there is the GPS status (A=active, V=void)
        GpsStatus = NmeaSentence[k + 1];
        break;
      }
    }
  }
  
  UpdateDisplay();
  ni = 0;
}


void InitDisplay(){ // init display
  tft.fillScreen(CYAN);
  tft.setCursor(1, 1);
  tft.setTextColor(BLACK);  
  tft.setTextSize(2);
  tft.println("  Speed");
  tft.setCursor(1, 59);
  tft.println("  G Force");
  tft.setCursor(1, 101);
  tft.println("  Acceleration");
}


void UpdateDisplay(){ // show the data on the display
  char DispBuffer[64];
  
  if (DispVel != Vel){ // update speed
    tft.fillRect(1,18,320,40,WHITE);
    tft.setCursor(1, 18);
    tft.setTextColor(BLACK);  
    tft.setTextSize(5);
    dtostrf(Vel, 3, 1, DispBuffer);
    strcat(DispBuffer, " km/h");
    tft.println(DispBuffer);
    DispVel = Vel;
  }

  if (DispGForc != GForc){ // g force
    tft.fillRect(1,76,320,24,WHITE);
    tft.setCursor(1, 76);
    tft.setTextColor(BLACK);  
    tft.setTextSize(3);
    dtostrf(GForc, 3, 3, DispBuffer);
    strcat(DispBuffer, " G ");
    int l = strlen(DispBuffer);
    *(DispBuffer + l) = Direc;
    *(DispBuffer + l + 1) = 0;    
    tft.println(DispBuffer);
    DispGForc = GForc;
  }
  
  if (DispAccel != Accel){ // acceleration
    tft.fillRect(1,118,320,24,WHITE);
    tft.setCursor(1, 118);
    tft.setTextColor(BLACK);  
    tft.setTextSize(3);
    dtostrf(Accel, 3, 2, DispBuffer);
    strcat(DispBuffer, " m/s2");
    tft.println(DispBuffer);
    DispAccel = Accel;
  }

  if (DispMaxVel != MaxVel){ // max speed
    tft.fillRect(1,150,320,16,WHITE);
    tft.setCursor(1, 150);
    tft.setTextColor(BLACK);  
    tft.setTextSize(2);
    strcpy(DispBuffer, "Max speed: ");
    dtostrf(MaxVel, 3, 1, DispBuffer + 11);
    strcat(DispBuffer, " km/h");
    tft.println(DispBuffer);
    DispMaxVel = MaxVel;
  }
  if (DispMaxGForc != MaxGForc){ // max g force
    tft.fillRect(1,168,320,16,WHITE);
    tft.setCursor(1, 168);
    tft.setTextColor(BLACK);  
    tft.setTextSize(2);
    strcpy(DispBuffer, "Max G for: ");
    dtostrf(MaxGForc, 3, 3, DispBuffer + 11);
    strcat(DispBuffer, " G ");
    int l = strlen(DispBuffer);
    *(DispBuffer + l) = MaxGForcDir;
    *(DispBuffer + l + 1) = 0;        
    tft.println(DispBuffer);
    DispMaxGForc = MaxGForc;
  }
  if (DispMaxAccel != MaxAccel){ // max acceleration
    tft.fillRect(1,186,320,16,WHITE);
    tft.setCursor(1, 186);
    tft.setTextColor(BLACK);  
    tft.setTextSize(2);
    strcpy(DispBuffer, "Max accel: ");
    dtostrf(MaxAccel, 3, 2, DispBuffer + 11);
    strcat(DispBuffer, " m/s2");
    tft.println(DispBuffer);
    DispMaxAccel = MaxAccel;
  }

  if (DispMinAccel != MinAccel){ // min acceleration
    tft.fillRect(1,204,320,16,WHITE);
    tft.setCursor(1, 204);
    tft.setTextColor(BLACK);  
    tft.setTextSize(2);
    strcpy(DispBuffer, "Max decel: ");
    dtostrf(MinAccel, 3, 2, DispBuffer + 11);
    strcat(DispBuffer, " m/s2");
    tft.println(DispBuffer);
    DispMinAccel = MinAccel;
  }
  
  if (DispGpsStatus != GpsStatus){ // gps status
    tft.setCursor(1, 224);
    tft.setTextColor(BLACK);  
    tft.setTextSize(2);
    strcpy(DispBuffer, "GPS status: ");
    if (GpsStatus == 'A'){
      tft.fillRect(0,222,320,18,GREEN);
      strcat(DispBuffer, "Active.");
    }else{
      tft.fillRect(0,222,320,18,RED);
      strcat(DispBuffer, "Seeking sat...");
      
    }
    tft.println(DispBuffer);
    DispGpsStatus = GpsStatus;
    
  }
}
